import json, string
import pandas as pd

DATA_LABELS = {
    "Prefab": 0,
    "Computer-Aided Design": 1,
    "Computer-Controlled Cutting": 2,
    "Embedded Programing": 3,
    "3D Scanning and Printing": 4,
    "Electronics Design": 5,
    "Computer-Controlled Machining": 6,
    "Electronics Production": 7,
    "Mechanical Design, Machine Design": 8,
    "Input Devices": 9,
    "Moulding and Casting": 10,
    "Output Devices": 11,
    "Embedded Networking and Communications": 12,
    "Interface and Application Programming": 13,
    "Wildcard Week": 14,
    "Applications and Implications": 15,
    "Invention, Intellectual Property and Business Models": 16,
    "Final Project": 17,
    "Other": 18
}

if __name__ == "__main__":
    txts = []
    labels = []
    tos = []
    froms = []
    with open('NLP_data/train.jsonl', 'r') as jsonl:
        for line in jsonl.readlines():
            j = json.loads(line)
            txts.append("".join([char for char in j['text'] if char in set(string.printable)])) # string.printable removes non-ASCII chars to avoid writing errors
            labels.append(DATA_LABELS[j['label']])
            tos.append(j['metadata']['to'])
            froms.append(j['metadata']['from'])

    data_dict = {
        "Text": txts,
        "Labels": labels,
        "To": tos,
        "From": froms
    }
    
    df = pd.DataFrame(data_dict)

    df.to_csv("NLP_data/train.csv")
```

**train_nn.py**

```py
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import CountVectorizer
import torch
import torch.nn as nn
import torch.optim as optim
from sklearn.metrics import accuracy_score
import pickle

# Read the CSV file
df = pd.read_csv('NLP_data/train.csv')

# Columns are named 'Text' and 'Labels'
texts = df['Text']
labels = df['Labels']

# Split the dataset into training and test sets
X_train, X_test, y_train, y_test = train_test_split(texts, labels, test_size=0.2, random_state=42)

# Convert the texts into vectors
vectorizer = CountVectorizer(max_features=5000)  # limit to 5000 most frequent words
X_train = vectorizer.fit_transform(X_train)
X_test = vectorizer.transform(X_test)

# Save the vectorizer
pickle.dump(vectorizer, open("vectorizer.pickle", "wb"))

# Convert the vectors and labels to numpy arrays
X_train = X_train.toarray()
X_test = X_test.toarray()
y_train = y_train.to_numpy()
y_test = y_test.to_numpy()

X_train = torch.from_numpy(X_train).float()
X_test = torch.from_numpy(X_test).float()
y_train = torch.from_numpy(y_train).long()
y_test = torch.from_numpy(y_test).long()

class TextClassifier(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim, dropout_prob=0.5):
        super(TextClassifier, self).__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.dropout = nn.Dropout(dropout_prob)
        self.fc2 = nn.Linear(hidden_dim, output_dim)
    
    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.fc2(x)
        return x

# Define your grid of hyperparameters
learning_rates = [0.1, 0.01, 0.001]
dropout_rates = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

# Define a function for your training loop
def train_model(lr, dropout_rate):
    model = TextClassifier(input_dim=X_train.shape[1], hidden_dim=200, output_dim=18, dropout_prob=dropout_rate)
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=lr)

    for epoch in range(100):  # you can adjust this number
        optimizer.zero_grad()
        outputs = model(X_train)
        loss = criterion(outputs, y_train)
        loss.backward()
        optimizer.step()

    # Evaluate the model
    with torch.no_grad():
        outputs = model(X_test)
        _, predicted = torch.max(outputs, 1)
        correct = (predicted == y_test).sum().item()
        accuracy = correct / y_test.size(0)
    
    return accuracy, model

# Perform the grid search
best_accuracy = 0.0
best_lr = None
best_dropout_rate = None
best_model = None

for lr in learning_rates:
    for dropout_rate in dropout_rates:
        accuracy, model = train_model(lr, dropout_rate)
        print(f'Learning rate: {lr}, Dropout rate: {dropout_rate}, Accuracy: {accuracy}')
        
        if accuracy > best_accuracy:
            best_accuracy = accuracy
            best_lr = lr
            best_dropout_rate = dropout_rate
            best_model = model

print(f'Best learning rate: {best_lr}, Best dropout rate: {best_dropout_rate}, Best accuracy: {best_accuracy}')

# Save the best model to a file
torch.save(best_model.state_dict(), 'best_model.pt')

# Also save the parameters in a dictionary
best_params = {"learning_rate": best_lr, "dropout_rate": best_dropout_rate, "accuracy": best_accuracy}

with open('best_params.pickle', 'wb') as handle:
    pickle.dump(best_params, handle, protocol=pickle.HIGHEST_PROTOCOL)