import requests, base64, urllib.parse, gitlab, re, pickle, os, time, json
from bs4 import BeautifulSoup
from loading_animations import start_loading, stop_loading
import pandas as pd
from urllib.parse import urljoin
from load_model_and_classify import Classifier

GL = gitlab.Gitlab('https://gitlab.fabcloud.org', api_version=4)
VALID_EXTENSOINS = ["html", "txt", "md"]
#_2023_SUBGROUP_ID = 8145
ALL_LAB_SUBGROUP_IDS = {
    2023: 8145,
    2022: 3632,
    2021: 2917,
    2020: 2140,
    2019: 1619,
    2018: 852
}

CHARACTERS_EACH_DIRECTION_TOPIC_DETECTION = 1000
TOPICS = [
    "Prefab",
    "Computer-Aided Design",
    "Computer-Controlled Cutting",
    "Embedded Programing",
    "3D Scanning and Printing",
    "Electronics Design",
    "Computer-Controlled Machining",
    "Electronics Production",
    "Mechanical Design, Machine Design",
    "Input Devices",
    "Moulding and Casting",
    "Output Devices",
    "Embedded Networking and Communications",
    "Interface and Application Programming",
    "Wildcard Week",
    "Applications and Implications",
    "Invention, Intellectual Property and Business Models",
    "Final Project",
    "Other"
]

TOPIC_SEARCH_TERMS = [
    [],
    ["Computer-Aided Design", "freecad"], 
    ["Laser Cut", "CCC week", "Computer-Controlled Cutting"],
    ["Embedded Programming", "MicroPython", "C\+\+", "python" "cpp", "ino", "Arduino IDE", "programming week"], 
    ["3D Printing", "3D Scanning", "TPU", "PETG", "filament", "Prusa", "3d printing week", "polyCAM", "3D Scanning and Printing", "3d printers", "3d printer"], # not PLA because in too many other words
    ["EagleCAD", "Eagle", "KiCAD", "Routing", "Auto-Route", "Trace", "Footprint", "electronic design", "Electronics Design"],
    ["CNC", "Shopbot", "Computer-Controlled Machining"],
    ["Mill", "Milling", "copper", "electronic production", "Electronics Production"],
    ["Machine week", "actuation and automation", "Mechanical Design, Machine Design"],
    ["Input Devices", "Input Device", "Inputs Devices", "Electronic input", "sensor", "Input Devices"],
    ["Part A", "Part B", "pot time", "pottime", "molding", "moulding", "casting", "cast", "Moulding and Casting"],
    ["Output Device", "Outputs Devices", "Outputs Device", "Servo", "motor", "Output Devices"],
    ["SPI", "UART", "I2C", "RX", "TX", "SCL", "networking week", "networking", "network", "networking and communications", "Embedded Networking and Communications"],
    ["Interfacing Week", "interface week", "Interface and Application Programming"],
    ["Wildcard Week"],
    ["Applications and Implications", "Bill-of-Materials", "Bill of materials"],
    ["Patent", "copyright", "trademark", "Invention, Intellectual Property and Business Models"],
    ["Final Project"],
    []

] # more than two times, lower(), separated or not by spaces or another character, fuzzy match to remove 's's, would probably make more sense to sort the sites then go by words; but this is fine for now

def get_file_content(file_path, project_id, default_branch_name):
    print(f"Getting file content: {file_path}, {project_id}")
    safe_url = f"https://gitlab.fabcloud.org/api/v4/projects/{project_id}/repository/files/{urllib.parse.quote(file_path, safe='')}?ref={default_branch_name}"
    print(safe_url)
    response = requests.get(safe_url).json()
    if 'message' in response:
        if '404' in response['message']:
            print(f"404 ERROR from {safe_url}")
            return
    encrypted_blob = response['content']
    
    decrypted_text = base64.b64decode(encrypted_blob)
    return decrypted_text

def get_repo_name(project_id):
    safe_url = f"https://gitlab.fabcloud.org/api/v4/projects/{project_id}"
    print(safe_url)
    response = requests.get(safe_url).json()
    web_url = response['web_url']
    if 'message' in response:
        if '404' in response['message']:
            print(f"404 ERROR from {safe_url}")
            return
    return response['name'], gitlab_url_to_site_url(f"{web_url}/") # / to make it consistent with the name lists with urljoin and the a's hrefs

def get_file_repo_list(id):
    all_file_paths = []
    project = GL.projects.get(id)#(f'academany/fabacademy/2023/labs/{lab}/students/{name}')
    all_directories = project.repository_tree(recursive=True, all=True)
    for item in all_directories:
        path = item['path']
        if path.split('.')[-1].lower().strip() in VALID_EXTENSOINS:
            all_file_paths.append(path)
    return all_file_paths

def get_subgroup_ids(group_id):
    safe_url = f"https://gitlab.fabcloud.org/api/v4/groups/{group_id}/subgroups"
    response = None
    while response is None:
        try:
            response = requests.get(safe_url).json()
        except Exception as e:
            print(f"Server timeout- {e}")
            time.sleep(10)
    p_ids = []
    for item in response:
        try:
            p_ids.append(item['id'])
        except:
            print("Error finding subgroups -- skipping")
    return p_ids

def get_subproject_ids(group_id):
    safe_url = f"https://gitlab.fabcloud.org/api/v4/groups/{group_id}/projects"
    response = requests.get(safe_url).json()
    p_ids = []
    for item in response:
        p_ids.append(item['id'])
    return p_ids

def name_split_char(year):
    return "." if year < 2021 else " "

def gitlab_url_to_site_url(gitlab_url):
    return f"https://fabacademy.org/{gitlab_url.split('https://gitlab.fabcloud.org/academany/fabacademy/')[-1]}"

def format_name(name_url_tup, year, tup=True):
    if tup:
        name, web_url = name_url_tup
        return f'{"-".join(name.lower().strip().split(name_split_char(year)))};{web_url}' 
    else:
        print("Warning: name generated without URL")
        name = name_url_tup
        return "-".join(name.lower().strip().split(name_split_char(year)))

def filter_only_student_repos(all_student_repo_ids, all_student_names, year):
    filtered_ids = []
    all_student_urls = [name.split(";")[1].strip() for name in all_student_names]
    for i, id_list in all_student_repo_ids:
        for id in id_list:
            _, web_url = get_repo_name(id)
            if web_url.strip()[:-1] in all_student_urls or web_url.strip() in all_student_urls: # [:-1] to remove ending slash
                filtered_ids.append((i, id))
    return filtered_ids

def get_all_student_repo_ids(year, year_subgroup_id, all_student_names):
    if save_exists("student_repo_id_saves", year):
        return load_obj("student_repo_id_saves", year)
    all_student_repo_ids = []
    all_lab_ids = get_subgroup_ids(year_subgroup_id)
    
    for id in all_lab_ids:
        for sub_id in get_subgroup_ids(id):
            all_student_repo_ids.append((id, get_subproject_ids(sub_id))) # ((i, get_subproject_ids(sub_id)))
    print(all_student_repo_ids)
    to_return = filter_only_student_repos(all_student_repo_ids, all_student_names, year)
    save_obj("student_repo_id_saves", to_return, year)
    return to_return

def save_exists(folder_name, name):
    return os.path.exists(f"{folder_name}/{name}.obj")

def load_obj(folder_name, name):
    with open(f"{folder_name}/{name}.obj", "rb") as filehandler:
        return pickle.load(filehandler)

def save_obj(folder_name, obj, name):
    with open(f"{folder_name}/{name}.obj", 'wb') as filehandler: 
        pickle.dump(obj, filehandler)

def get_all_people(year):
    if save_exists("people_saves", year):
        return load_obj("people_saves", year)
    
    base_url = f"https://fabacademy.org/{year}/people.html"
    soup = BeautifulSoup(requests.get(base_url).content, 'html.parser')

    if year > 2018:
        lab_divs = soup.find_all("div", {"class": "lab"})

        names = []

        for lab_div in lab_divs:
            lis = lab_div.find_all("li")

            As = [li.find("a") for li in lis]

            names += [f"{a.text.strip().lower().replace(' ', '-')};{urljoin(base_url, a['href'])}" for a in As]        
    else:
        lis = soup.find_all("li")
        names = [f"{li.find('a').text.strip().lower().replace(' ', '-')};{urljoin(base_url, li.find('a')['href'])}" for li in lis]

    save_obj("people_saves", names, year)

    return names

def get_references(content, year, from_url):
    pattern = re.compile(f"\/{year}\/labs\/[^\/]+\/students\/(\w|-)+\/")
    people_linked = {}
    for match in pattern.finditer(str(content)):
        full_url = f"https://fabacademy.org{match.group(0)}"
        print(f"FULL URL {full_url}")
        person = format_name((match.group(0).split("/")[-2], full_url), year)
        """if person in people_linked:
            people_linked[person] += 1
        else:
            people_linked[person] = 1"""
        

        link_label = None

        topic_search_start_ind = match.start() - CHARACTERS_EACH_DIRECTION_TOPIC_DETECTION
        if topic_search_start_ind < 0:
            topic_search_start_ind = 0
        topic_search_end_ind = match.end() + CHARACTERS_EACH_DIRECTION_TOPIC_DETECTION
        if topic_search_end_ind > len(content):
            topic_search_end_ind = len(content)
        print(f"({match.start()}, {match.end()}) -> ({topic_search_start_ind}, {topic_search_end_ind})")
        topic_text = content.decode()[topic_search_start_ind:topic_search_end_ind].lower()
        
        print("TOPIC TEXT", topic_text)
        for i in range(len(TOPICS)):
            topic = TOPICS[i]
            topic_search_terms = TOPIC_SEARCH_TERMS[i]
            for item in topic_search_terms:
                item_spaces = item.replace("-", " ").replace("/", " ").replace(",", " ").lower().strip()
                if re.search(re.compile(item_spaces.replace(" ",".")), topic_text) or re.search(re.compile(item_spaces.replace(" ","-")), topic_text) or re.search(re.compile(item_spaces), topic_text) or re.search(re.compile(item_spaces.replace(" ","")), topic_text):
                    link_label = topic
        
        if link_label is None:
            continue
            #print("No match so classifying as",end=" ")
            #link_label = TOPICS[classifier.classifyItem(content)]
            #print(f"{link_label}:")
            #print(topic_text)
            """ This code if still writing data to JSONL
            with open("sort.txt", "w") as file:
                file.write(topic_text)
            for i in range(len(TOPICS)):
                print(f"{i}: {TOPICS[i]}")
            ans = None
            while ans is None:
                try:
                    ans = int(input("Topic? "))
                except:
                    print("Error")

            link_label = TOPICS[ans]
            """
        
        # it's ok if its from someone to themselves ; uncomment this and add continue after previous if when not collecting data
        with open("NLP_data/train.jsonl", "a") as file:
            file.write('{' + '"text": ' + json.dumps(topic_text) + ', "label": "' + link_label + '", "metadata": '+ '{' + '"from": "' + from_url + '", "to": "' + full_url + '"' + '}' + '}\n')
        
        if person in people_linked:
            if link_label in people_linked[person]:
                people_linked[person][link_label] += 1
        else:
            people_linked[person] = {}
            people_linked[person][link_label] = 1

    return people_linked

def combine_reference_dicts(reference_dict_list):
    combined_reference_dict = {}
    for dict in reference_dict_list:
        for key in dict:
            if key in combined_reference_dict:
                for topic_key in dict[key]:
                    if topic_key in combined_reference_dict[key]:
                        combined_reference_dict[key][topic_key] += dict[key][topic_key]
                    else:
                        combined_reference_dict[key][topic_key] = dict[key][topic_key]
            else:
                combined_reference_dict[key] = dict[key] # might be a reference not a deepcopy but I don't think this should be an issue
    return combined_reference_dict

    """combined_reference_dict = {}
    for dict in reference_dict_list:
        for key in dict:
            if key not in combined_reference_dict:
                combined_reference_dict[key] = 1
            else:
                combined_reference_dict[key] += 1
    return combined_reference_dict"""

def format_data_to_matrix(data): # data

    #print("INCOMING DATA " + str(data))

    # OLD: data_test = [[(0, {"guy": {"adam": 1, "billy": 2}}), (0, {"adam": {"billy": 17, "guy": 1}}), (1, {"billy": {"adam": 3}})], [(2, {"new guy": {"billy": 3}})]]

    # data_test = [[(0, {"guy": {"adam": {"Final Project": 2, "Computer-Aided Design": 3}, "billy": {"Computer-Controlled Cutting" :4}}}), (0, {"adam": {"billy": {"Computer-Aided Design": 1}, "guy": {"Final Project": 1, "Prefab": 2}}}), (1, {"billy": {"adam": {"Prefab": 1}}})], [(2, {"new guy": {"billy": {"Prefab": 2}}})]]

    #df = pd.read_csv("data.csv", header=[0, 1], index_col=0)
    #print(df)
    #quit()

    students = []

    for year_info in data: # [(lab_id: int, {"Student Name": {"student-referenced": num_references: int, ...}}), ...]
        for student_info in year_info: # (lab_id: int, {"Student Name": {"student-referenced": num_references: int, ...}})
            lab_id = student_info[0] # oh no (not specifically this line) try to get student ids instead of student names to avoid overlap and later find site link
            student_name = list(student_info[1].keys())[0]
            reference_dict = student_info[1][list(student_info[1].keys())[0]]
            students.append(student_name) # would like to include info like ID, but it would be hard to do this from the link (matching a link to a unique student's repo without being able to get the link from gitlab (maybe?))

    print("STUDENTS", students)

    df = pd.crosstab(students, students)
    df.rename_axis("Referencing Students", axis=0, inplace=True)
    df.rename_axis("Referenced Students", axis=1, inplace=True)

    df = pd.DataFrame(df, index=df.index, columns=pd.MultiIndex.from_product([df.columns, TOPICS]))

    for student1 in students:
        for student2 in students:
            for topic_name in TOPICS:
                df.at[student1, (student2, topic_name)] = (pd.NA if student1 == student2 else 0)

    def assign_value(referencer_student, referenced_student, num_references, topic):
        if referencer_student == referenced_student:
            print(f"Ignoring self-referenced student {referencer_student}")
        elif referenced_student not in students:
            print("Referenced student isn't a student - URL matched naming convention so regex caught but wasn't checked against student list - skipping")
        else:
            #df.at[referencer_student, referenced_student, topic] = num_references
            #df.loc[referencer_student][referenced_student, topic] = num_references
            #df.insert(referencer_student, (referenced_student, topic), num_references)
            df.loc[referencer_student, (referenced_student, topic)] = num_references

    def get_value(referencer_student, referenced_student, topic):
        #return df.at[referencer_student, referenced_student, topic]
        return df.loc[referencer_student, (referenced_student, topic)]

    for year_info in data:
        for student_info in year_info:
            student_name = list(student_info[1].keys())[0]
            reference_dict = dict(student_info[1][student_name])
            for referenced_student in reference_dict:
                for reference_type in reference_dict[referenced_student]:
                    num_references = reference_dict[referenced_student][reference_type]
                    assign_value(student_name, referenced_student, num_references, reference_type)

    df.to_csv("final_data.csv", index_label="Referencing Students|Referenced Students")

    return df

def get_all_reference_dicts(year, id):
    filename = f"{year}-{id}"
    if save_exists("reference_dict_saves", filename):
        print(f"Save exists! {load_obj('reference_dict_saves', filename)}")
        return load_obj("reference_dict_saves", filename)
    reference_dict_list = []
    default_branch_name_response = requests.get(f"https://gitlab.fabcloud.org/api/v4/projects/{id}/repository/branches").json()
    default_branch_name = None
    for branch in default_branch_name_response:
        if branch['default']:
            default_branch_name = branch['name']
    if default_branch_name is None:
        print(f"Error: Default Branch Not Found (id: {id}) {default_branch_name_response} - leaving as None")
    try:
        for file in get_file_repo_list(id):
            print(f"Checking {file}...")
            reference_dict_list.append(get_references(get_file_content(file, id, default_branch_name), year, get_repo_name(id)[1]))
        print("Generating compiled reference dictionary...")
        compiled_reference_dict = combine_reference_dicts(reference_dict_list)
        print(compiled_reference_dict)
        save_obj("reference_dict_saves", compiled_reference_dict, filename)
        print(f"SAVING to {filename}")
        return compiled_reference_dict
    except gitlab.exceptions.GitlabGetError as e:
        return combine_reference_dicts(reference_dict_list)

if __name__ == "__main__":

    while True:
        try:
            classifier = Classifier()

            reference_dicts_across_years = [] # [[(lab_id: int, {"Student Name": {"student-referenced": num_references: int, ...}}), ...], ...]

            for year in range(2018, 2024):
                print("Loading student names...")
                all_student_names = get_all_people(year)
                print(all_student_names)
                print("Collecting student repo IDs...")
                all_student_repo_ids = get_all_student_repo_ids(year, ALL_LAB_SUBGROUP_IDS[year], all_student_names)
                print(all_student_repo_ids)

                all_reference_dicts = [] # [(lab_id: int, {"Student Name": {"student-referenced": num_references: int, ...}}), ...]

                #temp_i = 0

                for lab_number, id in all_student_repo_ids:
                    """temp_i += 1
                    if temp_i > 3:
                        break"""
                    reference_dict_list = []
                    
                    compiled_reference_dict = get_all_reference_dicts(year, id)

                    print("Adding reference dictionary to database...")
                    all_reference_dicts.append((lab_number, {format_name(get_repo_name(id), year): compiled_reference_dict}))
                    print(f"All reference dictionaries so far... {all_reference_dicts}")
                reference_dicts_across_years.append(all_reference_dicts)


            matrix = format_data_to_matrix(reference_dicts_across_years)
        except KeyboardInterrupt:
            quit()
        except Exception as e:
            print(e) 
            #pass # to try again for HTTP errors
        else:
            quit()
    


"""

DATA WRANGLING CONCERNS

SOLVED: Students like lucio-pentagnaneto:https://fabacademy.org/2018/labs/farmlabalgarve/students/lucio-pentagnaneto (2018) have different spaces in the url than in their name listed, so they are not recognized as students as their name does not match this list: https://fabacademy.org/2018/people.html
To fix this, I changed the filtering students code to only compare the URL

"""